/*

  This file is a part of MeanWalker, a library which provides utilities to
  sample from probability distributions using methods like the 
  Metropilis-Hastings algorithm and Goodman-Weare algorithm.

  Copyright (C) 2012 Jori Liesenborgs

  Contact: jori.liesenborgs@gmail.com
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
*/

#include "goodmanwearewalkers.h"
#include "randomnumbergenerator.h"
#include <cmath>

#define GOODMANWEAREWALKER_PI 3.14159265358979323846264338327950288419716939937510582097494459230

namespace meanwalker
{

GoodmanWeareWalkers::GoodmanWeareWalkers(RandomNumberGenerator *pRndGen) : Walker(pRndGen)
{
}

GoodmanWeareWalkers::~GoodmanWeareWalkers()
{
}

bool GoodmanWeareWalkers::init(Function *pProbFunction, const double *pStartCoordCenter, double startCoordRadius,
		               double a, int Nwalkers)
{
	int dim = pProbFunction->getDimension();
	std::vector<double> radii(dim);

	for (int i = 0 ; i < dim ; i++)
		radii[i] = startCoordRadius;

	return init(pProbFunction, pStartCoordCenter, &(radii[0]), a, Nwalkers);
}

bool GoodmanWeareWalkers::init(Function *pProbFunction, const double *pStartCoordCenter, const double *pStartCoordRadii,
		               double a, int Nwalkers)
{
	if (getProbabilityFunction() != 0)
	{
		setErrorString("Already initialized");
		return false;
	}
	if (pProbFunction == 0)
	{
		setErrorString("You need to specify a probability function");
		return false;
	}

	setProbabilityFunction(pProbFunction);
	
	int dim = pProbFunction->getDimension();
	RandomNumberGenerator *pRndGen = getRandomNumberGenerator();
	Function &probFunction = *(getProbabilityFunction());

	if (Nwalkers <= 0)
		Nwalkers = dim+1;

	if (Nwalkers <= dim)
	{
		setErrorString("The number of walkers must be at least one more as the dimension of the problem");
		return false;
	}

	if (a <= 1)
	{
		setErrorString("The 'a' parameter must be greater than one");
		return false;
	}

	m_bestCoords.resize(dim);
	m_tmp.resize(dim);

	m_a = a;
	m_aScale = std::sqrt(m_a) - 1.0/std::sqrt(m_a);
	m_aOffset = 1.0/std::sqrt(m_a);

	m_walkers.resize(Nwalkers);

	for (int l = 0 ; l < Nwalkers ; l++)
	{
		if (dim > 1)
		{
			// generate a random unit vector

			double sum = 0;

			for (int i = 0 ; i < dim ; i++)
			{
				double z = pRndGen->pickRandomNumber();
				double z2 = pRndGen->pickRandomNumber();
				double rho = -std::log(z);
				double theta = z2*2.0*GOODMANWEAREWALKER_PI;
				double x = std::sqrt(2.0*rho) * std::cos(theta);

				m_tmp[i] = x; // is a gaussian with sigma = 1

				sum += x*x;
			}
			sum = std::sqrt(sum);

			for (int i = 0 ; i < dim ; i++)
				m_tmp[i] /= sum;

			// use it to initialize a walker

			m_walkers[l].m_coords.resize(dim);

			for (int i = 0 ; i < dim ; i++)
				m_walkers[l].m_coords[i] = pStartCoordCenter[i] + pStartCoordRadii[i]*m_tmp[i];
		}
		else
		{
			double startCoordRadius = pStartCoordRadii[0];

			m_walkers[l].m_coords.resize(1);
			m_walkers[l].m_coords[0] = pStartCoordCenter[0] + startCoordRadius*(pRndGen->pickRandomNumber()-0.5)*2.0;
		}
		
		m_walkers[l].m_prob = probFunction(&(m_walkers[l].m_coords[0]));
	}

	m_bestCoords = m_walkers[0].m_coords;

	double bestProb = m_walkers[0].m_prob;

	for (int l = 1 ; l < Nwalkers ; l++)
	{
		if (m_walkers[l].m_prob > bestProb)
		{
			bestProb = m_walkers[l].m_prob;
			m_bestCoords = m_walkers[l].m_coords;
		}
	}

	m_nextStartPos = 0;

	return true;
}

void GoodmanWeareWalkers::walk(size_t numSteps)
{
	Function &probFunction = *(getProbabilityFunction());
	RandomNumberGenerator *pRndGen = getRandomNumberGenerator();
	bool isLog =  probFunction.isLog();
	size_t Nwalkers = m_walkers.size();
	size_t dim = m_bestCoords.size();
	double *pNewPos = &(m_tmp[0]);
	double bestProb = probFunction(&(m_bestCoords[0]));
	size_t s = 0;
	size_t startPos = m_nextStartPos;

	while (s < numSteps)
	{
		for (size_t l = startPos ; l < Nwalkers && s < numSteps ; l++, s++)
		{
			// pick a random walker
			
			size_t j = (size_t)(pRndGen->pickRandomNumber()*(double)(Nwalkers-1));

			if (j >= l)
				j++;

			double *pCur = &(m_walkers[l].m_coords[0]);
			double *pOther = &(m_walkers[j].m_coords[0]);

			// pick a Z value

			double F = pRndGen->pickRandomNumber();
			double Z = F * m_aScale + m_aOffset;

			Z *= Z;

			// Generate new trial

			for (size_t i = 0 ; i < dim ; i++)
				pNewPos[i] = pOther[i] + Z*(pCur[i]-pOther[i]);

			// Calc accept prob

			if (probFunction.isValid(pNewPos)) // make sure to check the validity of the coordinates before evaluating
			{
				double probNew = probFunction(pNewPos);
				double oldProb = m_walkers[l].m_prob;
				bool accept = false;
				
				if (!isLog)
				{
					double acceptProb = std::pow(Z, (double)(dim-1)) * probNew/oldProb;

					if (acceptProb >= 1.0)
						accept = true;
					else
					{
						if (pRndGen->pickRandomNumber() < acceptProb)
							accept = true;
					}
				}
				else
				{
					double logAcceptProb = (double)(dim-1)*std::log(Z) + probNew - oldProb;

					if (logAcceptProb >= 0)
						accept = true;
					else
					{
						if (pRndGen->pickRandomNumber() < std::exp(logAcceptProb))
							accept = true;
					}
				}

				// Update value

				if (accept)
				{
					for (size_t i = 0 ; i < dim ; i++)
						pCur[i] = pNewPos[i];

					m_walkers[l].m_prob = probNew;
				}

				if (probNew > bestProb)
				{
					bestProb = probNew;
					for (size_t i = 0 ; i < dim ; i++)
						m_bestCoords[i] = pNewPos[i];
				}
			}

			// call function for histograms etc...

			onNewSample(m_walkers[l].m_prob, &(m_walkers[l].m_coords[0]));
		}

		startPos = 0;
	}

	m_nextStartPos = (m_nextStartPos + numSteps)%Nwalkers;
}

} // end namespace   
