/*

  This file is a part of MeanWalker, a library which provides utilities to
  sample from probability distributions using methods like the 
  Metropilis-Hastings algorithm and Goodman-Weare algorithm.

  Copyright (C) 2012 Jori Liesenborgs

  Contact: jori.liesenborgs@gmail.com
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
*/

/**
 * \file histogrambuildertemplate.h
 */

#ifndef MEANWALKER_HISTOGRAMBUILDERTEMPLATE_H

#define MEANWALKER_HISTOGRAMBUILDERTEMPLATE_H

#include "meanwalkerconfig.h"
#include "randomnumbergenerator.h"
#include "histogram.h"
#include <vector>

namespace meanwalker
{

/** This is a template base class for the histogram builders.
 *  This is a template base class for the histogram builders. It contains all the histogram
 *  related code, so that the actual class only needs to provide the appropriate initialization
 *  for the type of walker used.
 */
template <class WalkerBase>
class MEANWALKER_IMPORTEXPORT HistogramBuilderTemplate : public WalkerBase
{
public:
	HistogramBuilderTemplate(RandomNumberGenerator *pRng) : WalkerBase(pRng)
	{
		m_init = false;
	}

	~HistogramBuilderTemplate()
	{
	}

	bool initNormal(const double *pMinCoords, const double *pMaxCoords, size_t dim, size_t numBins)
	{
		if (m_init)
		{
			errut::ErrorBase::setErrorString("Already initialized");
			return false;
		}

		m_delayed = false;
		m_init = true;
		m_numBins = numBins;

		m_histograms.resize(dim);

		for (size_t i = 0 ; i < m_histograms.size() ; i++)
		{
			m_histograms[i].setIndex(i);
			m_histograms[i].setInstance(this);
		}

		for (size_t i = 0 ; i < m_histograms.size() ; i++)
			m_histograms[i].init(pMinCoords[i], pMaxCoords[i], m_numBins);

		return true;
	}

	bool initDelayed(size_t dim, size_t numBins)
	{
		if (m_init)
		{
			errut::ErrorBase::setErrorString("Already initialized");
			return false;
		}
		m_delayed = true;
		m_init = true;
		m_numBins = numBins;

		m_histograms.resize(dim);

		for (size_t i = 0 ; i < m_histograms.size() ; i++)
		{
			m_histograms[i].setIndex(i);
			m_histograms[i].setInstance(this);
		}

		return true;
	}

	void destroy()
	{
		m_histograms.resize(0);
		m_init = false;
	}

	/** Let the walkers sample the probability distribution specified during initialization.
	 *  Let the walkers sample the probability distribution specified during initialization. During the
	 *  initial steps \c initSteps, the values are not processed into the histograms. During the final
	 *  steps \c finalSteps, the values are either processed into the histograms immediately (if the
	 *  boundaries for the histograms were specified during initialization), or are all recorded for
	 *  processing into histograms later. If the latter is the case, the HistogramBuilderTemplate::getHistogramMinMax function
	 *  is called before actually creating the histograms to determine their boundaries based on the
	 *  recorded values. This determination of the boundaries is only done the first time this
	 *  function is called and \c finalSteps is greater than zero.
	 */
	bool run(size_t initSteps, size_t finalSteps)
	{
		if (!m_init)
		{
			errut::ErrorBase::setErrorString("Not initialized");
			return false;
		}

		m_record = false; 
		WalkerBase::walk(initSteps); 

		if (finalSteps > 0)
		{
			if (m_delayed)
			{
				for (size_t i = 0 ; i < m_histograms.size() ; i++)
					m_histograms[i].reserveMemory(finalSteps);
			}

			m_record = true; 
			WalkerBase::walk(finalSteps); 

			if (m_delayed)
			{
				m_delayed = false;
				for (size_t i = 0 ; i < m_histograms.size() ; i++)
					m_histograms[i].processRecordedValues(m_numBins);

				std::vector<const double *> recordedValues(m_histograms.size());

				for (size_t i = 0 ; i < m_histograms.size() ; i++)
					recordedValues[i] = m_histograms[i].getRecordedValues();

				onRecordedValuesProcessed(&(recordedValues[0]), m_histograms.size(), m_histograms[0].recordSize());

				for (size_t i = 0 ; i < m_histograms.size() ; i++)
					m_histograms[i].clearRecordedValues();
			}
		}

		return true;
	}

	/** Clears the histogram contents (but does not reset the boundaries). */
	void clearHistograms()											
	{ 
		for (size_t i = 0 ; i < m_histograms.size() ; i++) 
			m_histograms[i].clear(); 
	}

	/** Obtain the histogram for coordinate \c i. */
	const Histogram &getHistogram(size_t i) const								
	{ 
		return m_histograms[i]; 
	}

	/** Obtain all histograms. */
	const std::vector<const Histogram *> getHistograms() const
	{
		std::vector<const Histogram *> h(m_histograms.size());

		for (size_t i = 0 ; i < h.size() ; i++)
			h[i] = &(m_histograms[i]);

		return h;
	}

	/** Print all histograms - or rather the corresponding probability density - to standard output. */
	void printHistograms() const
	{
		for (size_t i = 0 ; i < m_histograms.size() ; i++)
		{
			std::cout << "# Histogram " << i << std::endl;
			m_histograms[i].print();
		}
	}
protected:
	/** This is called on each new sample and by default it performs the necessary functions to process
	 *  the values in the histograms (so make sure you call this parent implementation when overriding the
	 *  function in a derived class). */
	void onNewSample(double functionValue, const double *pCoords)
	{
		if (m_record)
		{
			if (!m_delayed)
			{
				for (size_t i = 0 ; i < m_histograms.size() ; i++)
					m_histograms[i].process(pCoords[i]);
			}
			else
			{
				for (size_t i = 0 ; i < m_histograms.size() ; i++)
					m_histograms[i].record(pCoords[i]);
			}
		}
	}

	/** When calculating the histogram widths automatically this function is called to determine the
	 *  boundaries based on the recorded values.
	 *  When calculating the histogram widths automatically this function is called to determine the
	 *  boundaries based on the recorded values. By default, the boundaries are set to the minimum and
	 *  maximum values recorded.
	 *  \param idx Index of the coordinate for which the values were recorded.
	 *  \param pValues The recorded values.
	 *  \param numValues The number of recorded values.
	 *  \param minValue Set this to the minimum value you want to have histogram entries for.
	 *  \param maxValue Set this to the maximum value you want to have histogram entries for.
	 */
	virtual void getHistogramMinMax(size_t idx, const double *pValues, size_t numValues, double &minValue, double &maxValue)
	{
		DelayedHistogram::getDefaultHistogramMinMax(pValues, numValues, minValue, maxValue);
	}

	/** If delayed histogram creation is specified, this function will be called at the time that
	 *  the recorded samples are processed into the histograms.
	 *  If delayed histogram creation is specified, this function will be called at the time that
	 *  the recorded samples are processed into the histograms. The indexing for \c pValues is as
	 *  follows: pValues[histogramIndex][recordIndex] */
	virtual void onRecordedValuesProcessed(const double **pValues, size_t numHistograms, size_t numRecords)
	{ 
	}
private:
	class InternalDelayedHistogram : public DelayedHistogram
	{
	public:
		InternalDelayedHistogram()									{ m_pInstance = 0; m_idx = 0; }

		void setInstance(HistogramBuilderTemplate *pInst)						{ m_pInstance = pInst; }
		void setIndex(size_t idx)									{ m_idx = idx; }
	protected:
		void getHistogramMinMax(const double *pValues, size_t numValues, double &minValue, double &maxValue) { m_pInstance->getHistogramMinMax(m_idx, pValues, numValues, minValue, maxValue); }
	private:
		size_t m_idx;
		HistogramBuilderTemplate *m_pInstance;
	};

	std::vector<InternalDelayedHistogram> m_histograms;
	bool m_record, m_delayed, m_init;
	size_t m_numBins;
};

} // end namespace   

#endif // MEANWALKER_HISTOGRAMBUILDERTEMPLATE_H
