/*

  This file is a part of MeanWalker, a library which provides utilities to
  sample from probability distributions using methods like the 
  Metropilis-Hastings algorithm and Goodman-Weare algorithm.

  Copyright (C) 2012 Jori Liesenborgs

  Contact: jori.liesenborgs@gmail.com
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
*/

/**
 * \file metropolishistogrambuilder.h
 */

#ifndef MEANWALKER_METROPOLISHISTOGRAMBUILDER_H

#define MEANWALKER_METROPOLISHISTOGRAMBUILDER_H

#include "meanwalkerconfig.h"
#include "histogrambuildertemplate.h"
#include "metropoliswalker.h"
#include "histogram.h"

namespace meanwalker
{

/** This class uses the Metropolis-Hastings method to take samples from a probability
 *  distribution and automatically creates histograms for the marginalized
 *  distributions corresponding to each of the coordinates.
 *  This class uses the Metropolis-Hastings method to take samples from a probability
 *  distribution and automatically creates histograms for the marginalized
 *  distributions corresponding to each of the coordinates. It uses a MetropolisWalker
 *  instance to take the samples.
 */
class MEANWALKER_IMPORTEXPORT MetropolisHistogramBuilder : private HistogramBuilderTemplate<MetropolisWalker>
{
public: 
	/** The constructor of the class, in which the random number generator to be used
	 *  must be specified. */
	MetropolisHistogramBuilder(RandomNumberGenerator *pRndGen);
	~MetropolisHistogramBuilder();

	/** Initialize the builder.
	 *  Initialize the builder.
	 *  \param pProbFunction The probability density to be sampled.
	 *  \param pMinCoords Sets the minimum coordinates which are used together with the maximum coordinates
	 *                    to calculate the scale of the steps used in the random walk.
	 *  \param pMaxCoords Sets the maximum coordinates which are used together with the minimum coordinates
	 *                    to calculate the scale of the steps used in the random walk.
	 *  \param pStartCoords If not null, the walker will start at these coordinates. If null, a random position
	 *                    between \c pMinCoords and \c pMaxCoords is used as starting point.
	 *  \param numBins Number of bins the histograms will have.
	 *  \param keepInside If true, the random walk is not allowed to go beyond the boundaries specified by
	 *                    the minimum coordinates. If false, the walker can go outside these bounds. In that
	 *                    case the boundaries are only used to calculate the step size.
	 *  \param delayed If false, the minimum and maximum coordinates will be used as the histogram boundaries.
	 *                 If true, the widths of the histograms will be determined automatically.
	 */
	bool init(Function *pProbFunction, const double *pMinCoords, const double *pMaxCoords, const double *pStartCoords,
		  size_t numBins, bool keepInside = true, bool delayed = false);

	using HistogramBuilderTemplate<MetropolisWalker>::run;
	using HistogramBuilderTemplate<MetropolisWalker>::clearHistograms;
	using HistogramBuilderTemplate<MetropolisWalker>::getHistogram;
	using HistogramBuilderTemplate<MetropolisWalker>::getHistograms;
	using HistogramBuilderTemplate<MetropolisWalker>::printHistograms;
	using MetropolisWalker::getDimension;
	using MetropolisWalker::getBestCoordinates;

	using ErrorBase::getErrorString;
protected:
	using HistogramBuilderTemplate<MetropolisWalker>::onNewSample;
	using HistogramBuilderTemplate<MetropolisWalker>::getHistogramMinMax;
	using HistogramBuilderTemplate<MetropolisWalker>::onRecordedValuesProcessed;
	using MetropolisWalker::stepToNewCoords;
};

} // end namespace   

#endif // MEANWALKER_METROPOLISHISTOGRAMBUILDER_H

